"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var _a;
var os = __importStar(require("os"));
var ISerializeSymbol_1 = require("./ISerializeSymbol");
var TYPES = (_a = {},
    _a[ISerializeSymbol_1.MODIFIER_TYPE.PRIVATE] = '-',
    _a[ISerializeSymbol_1.MODIFIER_TYPE.PROTECTED] = '#',
    _a[ISerializeSymbol_1.MODIFIER_TYPE.PUBLIC] = '+',
    _a);
var COMPOSITION_LINE = '*--';
function convertToPlant(tsjs, options) {
    if (options === void 0) { options = {
        compositions: false,
        onlyInterfaces: false
    }; }
    var lines = [];
    var compositions = [];
    var listOfSerializeSymbols = tsjs;
    if (options.onlyInterfaces) {
        listOfSerializeSymbols = listOfSerializeSymbols.filter(function (serializedSymbol) { return serializedSymbol.structure === ISerializeSymbol_1.STRUCTURE.INTERFACE; });
    }
    lines.push('@startuml');
    listOfSerializeSymbols.forEach(function (serializedSymbol) {
        var keyword = '';
        if (serializedSymbol.keyword !== undefined) {
            keyword = serializedSymbol.keyword + " ";
        }
        var heritage = '';
        if (serializedSymbol.extends !== undefined) {
            heritage += " extends " + serializedSymbol.extends;
        }
        var serializedClass = serializedSymbol;
        if (serializedClass.implements !== undefined) {
            heritage += " implements " + serializedClass.implements.join(', ');
        }
        var parameters = '';
        if (serializedClass.parameters !== undefined && serializedClass.parameters.length > 0) {
            parameters = "<" + serializedClass.parameters.map(function (parameter) {
                if (parameter.constraint === undefined) {
                    return parameter.name;
                }
                return parameter.name + " extends " + parameter.constraint;
            })
                .join(', ') + ">";
        }
        var openingBrace = '';
        if (serializedSymbol.members.length > 0) {
            openingBrace = ' {';
        }
        lines.push("" + keyword + serializedSymbol.structure + " " + serializedSymbol.name + parameters + heritage + openingBrace);
        serializedSymbol.members.forEach(function (serializedMember) {
            checkCompositions(serializedMember, serializedSymbol.name);
            lines.push(memberToString(serializedMember, serializedSymbol.name));
        });
        if (serializedSymbol.members.length > 0) {
            lines.push('}');
        }
    });
    if (compositions.length > 0 && options.compositions) {
        var uniqueCompositions = compositions.filter(function (value, index, array) { return array.indexOf(value) === index; });
        uniqueCompositions.forEach(function (composition) { return lines.push(composition); });
    }
    lines.push('@enduml');
    return lines.join(os.EOL);
    function memberToString(member, parentName) {
        var line = '    ';
        if (member.type === ISerializeSymbol_1.MEMBER_TYPE.ENUM) {
            return "" + line + member.name;
        }
        line += TYPES[member.modifierType];
        if (member.keyword !== undefined) {
            line += "{" + member.keyword + "} ";
        }
        line += "" + member.name + (member.questionToken !== undefined && member.questionToken ? '?' : '');
        if (member.type === ISerializeSymbol_1.MEMBER_TYPE.METHOD) {
            line += "(" + member.parameters.map(function (parameter) {
                return "" + parameter.name + (parameter.questionToken !== undefined && parameter.questionToken ? '?' : '') + ": " + parameter.type;
            })
                .join(', ') + ")";
        }
        if (member.returnType !== undefined) {
            line += ": " + member.returnType;
        }
        return line;
    }
    function checkCompositions(member, parentName) {
        listOfSerializeSymbols.some(function (serializedSymbolToSearch) {
            if (parentName === serializedSymbolToSearch.name) {
                return false;
            }
            var memberTypeIndex = member.type.split(' | ')
                .indexOf(serializedSymbolToSearch.name);
            var memberReturnTypeIndex = member.returnType.split(' | ')
                .indexOf(serializedSymbolToSearch.name);
            var memberParameterTypes = [];
            member.parameters.forEach(function (parameter) { return memberParameterTypes.push.apply(memberParameterTypes, parameter.type.split(' | ')); });
            var memberParamterTypesIndex = memberParameterTypes.indexOf(serializedSymbolToSearch.name);
            if (memberTypeIndex < 0 && memberReturnTypeIndex < 0 && memberParamterTypesIndex < 0) {
                return false;
            }
            compositions.push(parentName + " " + COMPOSITION_LINE + " " + serializedSymbolToSearch.name);
            return true;
        });
    }
}
exports.convertToPlant = convertToPlant;
