"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var typescript_1 = __importDefault(require("typescript"));
var ISerializeSymbol_1 = require("./ISerializeSymbol");
function generateDocumentation(fileNames, options) {
    if (options === void 0) { options = {
        module: typescript_1.default.ModuleKind.CommonJS,
        target: typescript_1.default.ScriptTarget.ES2015
    }; }
    var program = typescript_1.default.createProgram(fileNames, options);
    var checker = program.getTypeChecker();
    var output = [];
    program.getSourceFiles()
        .forEach(function (sourceFile) {
        if (!sourceFile.isDeclarationFile) {
            typescript_1.default.forEachChild(sourceFile, visit);
        }
    });
    return output;
    function visit(node) {
        if (!isNodeExported(node)) {
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.ClassDeclaration) {
            var currentNode = node;
            if (currentNode.name === undefined) {
                return;
            }
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            output.push(serializeClass(symbol));
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.InterfaceDeclaration) {
            var currentNode = node;
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            output.push(serializeInterface(symbol));
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.ModuleDeclaration) {
            typescript_1.default.forEachChild(node, visit);
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.EnumDeclaration) {
            var currentNode = node;
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            output.push(serializeEnum(symbol));
            return;
        }
    }
    function serializeSymbol(symbol) {
        var declarations = symbol.getDeclarations();
        var questionToken = false;
        if (declarations !== undefined) {
            questionToken = declarations.some(function (declaration) {
                if (declaration.questionToken === undefined &&
                    declaration.initializer === undefined) {
                    return false;
                }
                return true;
            });
        }
        return {
            name: symbol.getName(),
            type: checker.typeToString(checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration)),
            questionToken: questionToken
        };
    }
    function serializeMember(memberSymbol, memberDeclaration) {
        var result = {
            keyword: getMemberKeyword(memberDeclaration),
            modifierType: getMemberModifierType(memberDeclaration),
            name: memberSymbol.getName(),
            parameters: [],
            returnType: getMemberReturnType(memberSymbol, memberDeclaration),
            type: getMemberType(memberDeclaration),
            questionToken: getQuestionToken(memberDeclaration),
            constraint: getConstraint(memberDeclaration)
        };
        if (result.type === ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR) {
            return;
        }
        if (result.type === ISerializeSymbol_1.MEMBER_TYPE.METHOD) {
            var methodSignature = checker.getSignatureFromDeclaration(memberDeclaration);
            if (methodSignature === undefined) {
                return result;
            }
            return __assign({}, result, serializeSignature(methodSignature));
        }
        return result;
    }
    function getMemberReturnType(memberSymbol, memberDeclaration) {
        var memberType = getMemberType(memberDeclaration);
        if (memberType === ISerializeSymbol_1.MEMBER_TYPE.ENUM || memberType === ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR) {
            return '';
        }
        return checker.typeToString(checker.getTypeOfSymbolAtLocation(memberSymbol, memberDeclaration));
    }
    function getQuestionToken(memberDeclaration) {
        var propertyDeclaration = memberDeclaration;
        return propertyDeclaration.questionToken !== undefined;
    }
    function getConstraint(memberDeclaration) {
        var effectiveConstraint = typescript_1.default.getEffectiveConstraintOfTypeParameter(memberDeclaration);
        if (effectiveConstraint === undefined) {
            return;
        }
        return checker.typeToString(checker.getTypeFromTypeNode(effectiveConstraint));
    }
    function getMemberType(memberDeclaration) {
        switch (memberDeclaration.kind) {
            case typescript_1.default.SyntaxKind.PropertyDeclaration:
            case typescript_1.default.SyntaxKind.PropertySignature:
            case typescript_1.default.SyntaxKind.Parameter:
            case typescript_1.default.SyntaxKind.GetAccessor:
            case typescript_1.default.SyntaxKind.SetAccessor:
                return ISerializeSymbol_1.MEMBER_TYPE.PROPERTY;
            case typescript_1.default.SyntaxKind.MethodDeclaration:
            case typescript_1.default.SyntaxKind.MethodSignature:
                return ISerializeSymbol_1.MEMBER_TYPE.METHOD;
            case typescript_1.default.SyntaxKind.Constructor:
            case typescript_1.default.SyntaxKind.ConstructSignature:
                return ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR;
            case typescript_1.default.SyntaxKind.IndexSignature:
                return ISerializeSymbol_1.MEMBER_TYPE.INDEX;
            case typescript_1.default.SyntaxKind.TypeParameter:
                return ISerializeSymbol_1.MEMBER_TYPE.PARAMETER;
            case typescript_1.default.SyntaxKind.EnumMember:
            case typescript_1.default.SyntaxKind.EnumDeclaration:
                return ISerializeSymbol_1.MEMBER_TYPE.ENUM;
            default:
                throw new Error('unable to determine member type');
        }
    }
    function getModifierType(modifiers) {
        var modifierType = ISerializeSymbol_1.MODIFIER_TYPE.PUBLIC;
        modifiers.some(function (modifier) {
            if (modifier.kind === typescript_1.default.SyntaxKind.PrivateKeyword) {
                modifierType = ISerializeSymbol_1.MODIFIER_TYPE.PRIVATE;
                return true;
            }
            if (modifier.kind === typescript_1.default.SyntaxKind.PublicKeyword) {
                modifierType = ISerializeSymbol_1.MODIFIER_TYPE.PUBLIC;
                return true;
            }
            if (modifier.kind === typescript_1.default.SyntaxKind.ProtectedKeyword) {
                modifierType = ISerializeSymbol_1.MODIFIER_TYPE.PROTECTED;
                return true;
            }
            return false;
        });
        return modifierType;
    }
    function getMemberModifierType(memberDeclaration) {
        var memberModifiers = memberDeclaration.modifiers;
        if (memberModifiers === undefined ||
            memberModifiers.length === 0) {
            return ISerializeSymbol_1.MODIFIER_TYPE.PUBLIC;
        }
        return getModifierType(memberModifiers);
    }
    function getMemberKeyword(memberDeclaration) {
        var memberKeyword;
        var memberModifiers = memberDeclaration.modifiers;
        if (memberModifiers === undefined ||
            memberModifiers.length === 0) {
            return memberKeyword;
        }
        memberModifiers.some(function (memberModifier) {
            if (memberModifier.kind === typescript_1.default.SyntaxKind.AbstractKeyword) {
                memberKeyword = ISerializeSymbol_1.CLASS_MEMBER_KEYWORD.ABSTRACT;
                return true;
            }
            if (memberModifier.kind === typescript_1.default.SyntaxKind.StaticKeyword) {
                memberKeyword = ISerializeSymbol_1.CLASS_MEMBER_KEYWORD.STATIC;
                return true;
            }
            return false;
        });
        return memberKeyword;
    }
    function getExtendsHeritageClauseName(heritageClause) {
        return heritageClause.types[0].expression.text;
    }
    function getInterfaceName(nodeObject) {
        return nodeObject.expression.text;
    }
    function getImplementsHeritageClauseNames(heritageClause) {
        return heritageClause.types.map(getInterfaceName);
    }
    function serializeEnum(enumSymbol) {
        return __assign({}, serializeInterface(enumSymbol), { structure: ISerializeSymbol_1.STRUCTURE.ENUM });
    }
    function serializeInterface(interfaceSymbol) {
        var serializedInterface = __assign({}, serializeSymbol(interfaceSymbol), { members: [], structure: ISerializeSymbol_1.STRUCTURE.INTERFACE });
        if (interfaceSymbol.members !== undefined) {
            interfaceSymbol.members.forEach(handleInterfaceMemberSymbols);
        }
        if (interfaceSymbol.exports !== undefined) {
            interfaceSymbol.exports.forEach(handleInterfaceMemberSymbols);
        }
        if (interfaceSymbol.globalExports !== undefined) {
            interfaceSymbol.globalExports.forEach(handleInterfaceMemberSymbols);
        }
        var interfaceDeclarations = interfaceSymbol.getDeclarations();
        if (interfaceDeclarations === undefined) {
            return serializedInterface;
        }
        interfaceDeclarations.forEach(function (interfaceDeclaration) {
            var heritageClauses = interfaceDeclaration.heritageClauses;
            if (heritageClauses === undefined) {
                return;
            }
            heritageClauses.forEach(function (heritageClause) {
                if (heritageClause.token === typescript_1.default.SyntaxKind.ExtendsKeyword) {
                    serializedInterface.extends = getExtendsHeritageClauseName(heritageClause);
                    return;
                }
                if (heritageClause.token === typescript_1.default.SyntaxKind.ImplementsKeyword) {
                    serializedInterface.implements = getImplementsHeritageClauseNames(heritageClause);
                    return;
                }
                throw new Error('unsupported heritage clause');
            });
        });
        return serializedInterface;
        function handleInterfaceMemberSymbols(memberSymbol) {
            var memberDeclarations = memberSymbol.getDeclarations();
            if (memberDeclarations === undefined) {
                return;
            }
            memberDeclarations.forEach(function (memberDeclaration) {
                var serializedMember = serializeMember(memberSymbol, memberDeclaration);
                if (serializedMember === undefined) {
                    return;
                }
                if (serializedMember.type === ISerializeSymbol_1.MEMBER_TYPE.PARAMETER) {
                    if (serializedInterface.parameters === undefined) {
                        serializedInterface.parameters = [];
                    }
                    serializedInterface.parameters.push(serializedMember);
                    return;
                }
                serializedInterface.members.push(serializedMember);
            });
        }
    }
    function serializeClass(classSymbol) {
        var classKeyword;
        var classDeclaration = classSymbol.getDeclarations();
        if (classDeclaration !== undefined && classDeclaration.length > 0) {
            classKeyword = getMemberKeyword(classDeclaration[classDeclaration.length - 1]);
        }
        var constructorType = checker.getTypeOfSymbolAtLocation(classSymbol, classSymbol.valueDeclaration);
        return __assign({}, serializeInterface(classSymbol), { structure: ISerializeSymbol_1.STRUCTURE.CLASS, keyword: classKeyword, constructors: constructorType.getConstructSignatures()
                .map(serializeSignature) });
    }
    function serializeSignature(signature) {
        return {
            parameters: signature.parameters.map(serializeSymbol),
            returnType: checker.typeToString(signature.getReturnType())
        };
    }
    function isNodeExported(node) {
        return (node.flags & typescript_1.default.ModifierFlags.Export) !== 0 ||
            node.parent.kind === typescript_1.default.SyntaxKind.SourceFile;
    }
}
exports.generateDocumentation = generateDocumentation;
